from enum import Enum
from io import BytesIO
from typing import BinaryIO, List, Optional, Tuple


class Tool(Enum):
    NONE = 0
    SPOT = 1
    AREA = 2
    ELLIPSE = 3
    LINE = 4
    ENDPOINT = 5
    ALARM = 6
    UNUSED = 7
    DIFFERENCE = 8


# TODO Use specific Params types instead of list[int]
# SpotParams = tuple[int, int]
# AreaParams = tuple[int, int, int, int]
# EllipseParams = tuple[int, int, int, int, int, int]
# LineParams = tuple[int, int, int, int]
# Params = SpotParams | AreaParams | EllipseParams | LineParams | None


class Measurement:
    tool: Tool
    params: List[int]
    label: str

    def __init__(self, tool: Tool, params: List[int], label: str):
        self.tool = tool
        self.params = params
        self.label = label


def parse_measurements(
    stream: BinaryIO, metadata: Tuple[int, int, int, int]
) -> List[Measurement]:
    (_, _, offset, length) = metadata
    stream.seek(offset)
    measurements_bytes = stream.read(length)
    stream = BytesIO(measurements_bytes)
    stream.seek(12, 1)  # ignore first bytes

    measurements = []
    measurement_length = int.from_bytes(stream.read(1), "big") - 1
    while measurement_length > 0:
        measurement_bytes = BytesIO(stream.read(measurement_length))
        measurement = parse_measurement(measurement_bytes)
        if measurement:
            measurements.append(measurement)

        measurement_length = int.from_bytes(stream.read(1), "big") - 1
    return measurements


def parse_measurement(stream: BinaryIO) -> Optional[Measurement]:
    stream.seek(3, 1)
    num_bytes_params = int.from_bytes(stream.read(1), "big")
    num_bytes_label = int.from_bytes(stream.read(2), "big")

    stream.seek(3, 1)
    tool = int.from_bytes(stream.read(1), "big")
    # measurement_no = int.from_bytes(stream.read(2), "big")

    stream.seek(25, 1)
    params: List[int] = []
    while len(params) * 2 < num_bytes_params:
        param = int.from_bytes(stream.read(2), "little")
        params.append(param)

    label = stream.read(num_bytes_label).decode("utf-16le").rstrip("\x00")

    values = [v.value for v in Tool]
    return Measurement(Tool(tool), params, label) if tool in values else None



from io import BytesIO
from typing import BinaryIO, Dict, Tuple, Union, Callable
import struct

import numpy as np
from PIL import Image

from flyr.dtypes import RawArray


def parse_camera_info(
    stream: BinaryIO, metadata: Tuple[int, int, int, int]
) -> Dict[str, Union[int, float]]:
    (_, _, offset, _) = metadata
    stream.seek(offset + 32)

    emissivity = stream.read(4)
    object_distance = stream.read(4)
    refl_app_temp = stream.read(4)
    atmos_temp = stream.read(4)
    ir_window_temp = stream.read(4)
    ir_window_transm = stream.read(4)

    stream.seek(offset + 60)
    rel_humidity = stream.read(4)

    stream.seek(offset + 88)
    planck_r1 = stream.read(4)
    planck_b = stream.read(4)
    planck_f = stream.read(4)

    stream.seek(offset + 112)
    atmospheric_trans_alpha1 = stream.read(4)
    atmospheric_trans_alpha2 = stream.read(4)
    atmospheric_trans_beta1 = stream.read(4)
    atmospheric_trans_beta2 = stream.read(4)
    atmospheric_trans_x = stream.read(4)

    cam_temp_range_max = stream.read(4)
    cam_temp_range_min = stream.read(4)

    stream.seek(offset + 776)
    planck_o = stream.read(4)
    planck_r2 = stream.read(4)

    stream.seek(offset + 784)
    raw_value_range_min = stream.read(2)
    raw_value_range_max = stream.read(2)

    stream.seek(offset + 824)
    raw_value_median = stream.read(4)
    raw_value_range = stream.read(4)

    # Create result dictionary
    float_from_bytes = lambda v: struct.unpack("f", v)[0]  # noqa: E731
    camera_info = {
        "emissivity": float_from_bytes(emissivity),
        "object_distance": float_from_bytes(object_distance),
        "atmospheric_temperature": float_from_bytes(atmos_temp),
        "ir_window_temperature": float_from_bytes(ir_window_temp),
        "ir_window_transmission": float_from_bytes(ir_window_transm),
        "reflected_apparent_temperature": float_from_bytes(refl_app_temp),
        "relative_humidity": float_from_bytes(rel_humidity),
        "planck_r1": float_from_bytes(planck_r1),
        "planck_r2": float_from_bytes(planck_r2),
        "planck_b": float_from_bytes(planck_b),
        "planck_f": float_from_bytes(planck_f),
        "planck_o": struct.unpack("i", planck_o)[0],
        "atmospheric_trans_alpha1": float_from_bytes(atmospheric_trans_alpha1),
        "atmospheric_trans_alpha2": float_from_bytes(atmospheric_trans_alpha2),
        "atmospheric_trans_beta1": float_from_bytes(atmospheric_trans_beta1),
        "atmospheric_trans_beta2": float_from_bytes(atmospheric_trans_beta2),
        "atmospheric_trans_x": float_from_bytes(atmospheric_trans_x),
        "raw_value_range_min": struct.unpack("H", raw_value_range_min)[0],
        "raw_value_range_max": struct.unpack("H", raw_value_range_max)[0],
        "raw_value_median": struct.unpack("i", raw_value_median)[0],
        "raw_value_range": struct.unpack("i", raw_value_range)[0],
        "camera_temperature_range_max": float_from_bytes(cam_temp_range_max),
        "camera_temperature_range_min": float_from_bytes(cam_temp_range_min),
    }

    to_round = [
        "atmospheric_temperature",
        "ir_window_temperature",
        "reflected_apparent_temperature",
    ]
    camera_info = {k: round(v, 2) if k in to_round else v for k, v in camera_info.items()}

    return camera_info


def parse_raw_data(
    stream: BinaryIO, metadata: Tuple[int, int, int, int]
) -> Tuple[int, int, RawArray]:
    (_, _, offset, length) = metadata
    stream.seek(offset)

    # from PIL import ImageFile  # Uncomment to enable loading corrupt PNGs
    # ImageFile.LOAD_TRUNCATED_IMAGES = True

    stream.seek(2, 1)  # Skip first two bytes, TODO Explain the why of the skip
    width = int.from_bytes(stream.read(2), "little")
    height = int.from_bytes(stream.read(2), "little")

    # Read the bytes with the raw thermal data and decode using PIL
    length = min(width * height * 2, length)
    stream.seek(offset + 2 * 16)  # TODO document why 2 * 16
    thermal_bytes = stream.read(length)
    thermal_stream = BytesIO(thermal_bytes)

    # Some thermograms have the raw data stored as PNG files, others simply the
    # sequence of bytes
    thermal_np: RawArray
    if thermal_bytes[:4] != b"\x89PNG":
        u16s = []
        u16 = thermal_stream.read(2)
        while u16:
            u16s.append(int.from_bytes(u16, "little"))
            u16 = thermal_stream.read(2)
        thermal_np = np.array(u16s).reshape((height, width))
    else:
        # FLIR PNG data is in the wrong byte order
        thermal_img = Image.open(thermal_stream)
        thermal_np = np.array(thermal_img)
        fix_byte_order: Callable[[RawArray], RawArray]
        fix_byte_order = np.vectorize(lambda x: (x >> 8) + ((x & 0x00FF) << 8))
        thermal_np = fix_byte_order(thermal_np)

    # Check shape
    if thermal_np.shape != (height, width):
        msg = "Invalid FLIR: metadata's width and height don't match thermal data's actual width and height ({} vs ({}, {})"
        msg = msg.format(thermal_np.shape, height, width)
        raise ValueError(msg)

    return width, height, thermal_np



import os


def get() -> str:
    """Identify version number"""
    parent_dir = os.path.dirname(__file__)
    version_file = os.path.join(parent_dir, "VERSION.txt")
    assert os.path.isfile(version_file)
    with open(version_file) as fh:
        version = fh.read().strip()

    version_parts = version[1:].split(".")
    valid_version = (
        len(version) >= 5
        and version[0] == "v"
        and len(version_parts) >= 3
        and version_parts[0].isdigit()
        and version_parts[1].isdigit()
        and version_parts[2].isdigit()
    )
    assert valid_version

    return version.strip("v")



from typing import Literal, Union

import numpy as np

from flyr.dtypes import Generic1DArray, Generic2DArray

NormalizationMode = Union[Literal["minmax"], Literal["percentiles"], Literal["hist-eq"]]


def by_minmax(
    min_t: float, max_t: float, thermal: Generic2DArray[np.float64]
) -> Generic2DArray[np.float64]:
    """Normalize the given array to be between 0 and 1 using the given minimum
    and maximum values.

    Parameters
    ----------
    min_t: float
        The value to clip the input array to and which should map to 0.
    max_t: float
        The value to clip the input array to and which should map to 1.
    thermal: Generic2DArray[np.float64]
        The array to normalize.

    Returns
    -------
    Generic2DArray[np.float64]
        An array with all values between 0 and 1
    """
    err_msg = f"Minimum value {min_t} should be smaller than maximum value {max_t}"
    assert min_t < max_t, err_msg
    thermal = np.clip(thermal, min_t, max_t)
    thermal = (thermal - min_t) / (max_t - min_t)
    return thermal


def by_histogram_equalization(
    min_t: float,
    max_t: float,
    thermal: Generic2DArray[np.float64],
    num_colors: int = 256,
) -> Generic2DArray[np.float64]:
    """Normalize the given array using histogram equalization, improving
    contrast.

    Warning: Improves contrast, but changes the apparant temperature.

    Parameters
    ----------
    min_t: float
        The value to clip the input array to and which should map to 0.
    max_t: float
        The value to clip the input array to and which should map to 1.
    thermal: Generic2DArray[np.float64]
        The array to normalize.
    num_colors: int
        How many colors to equalize the histogram over. In other words, the
        number of histogram bins.

    Returns
    -------
    Generic2DArray[np.float64]
        An array with all values between 0 and 1
    """
    err_msg = f"Minimum value {min_t} should be smaller than maximum value {max_t}"
    assert min_t < max_t, err_msg

    normalized = by_minmax(min_t, max_t, thermal)
    return equalize_histogram(normalized, num_colors)


def equalize_histogram(normalized: Generic2DArray, num_colors: int):
    max_color_value = num_colors - 1
    color_chances, colored = __calc_color_chances(normalized, num_colors)
    new_colors = __calc_new_colors(color_chances)
    return new_colors[colored] / max_color_value


def __calc_color_chances(
    normalized: Generic2DArray[np.float64], num_colors: int
) -> tuple[Generic1DArray[np.float64], Generic2DArray[np.int64]]:
    """

    Parameters
    ----------
    normalized
        A 2D float numpy array of the thermal data normalized to between 0 and 1.
    num_colors: int
        The number of colors we want to calculate the occurrance chance of.

    Returns
    -------
    tuple[color chances, color map]
        A 2-tuple, where the first item is an array whose value indicates
        the chance to occur of the color identified by the index in this array.
        The second item is a map, derived from the normalized array, where
        the values correspond to color indices from the color chances array.
    """
    max_color = num_colors - 1
    color_map = (normalized * max_color).astype(int)

    [histogram, _] = np.histogram(color_map, bins=num_colors, range=(0, max_color))
    color_chances = histogram / normalized.size

    return color_chances, color_map


def __calc_new_colors(
    color_chances: Generic1DArray[np.float64],
) -> Generic2DArray[np.int64]:
    """Calculate a new color based on the current color.

    s = T(i) where i is a pixel's intensity
    0 <= i <= 255
    T(i) <= T(i+1)
    0 <= T(i) <= 255
    """
    max_color = color_chances.shape[0] - 1
    cdf = np.cumsum(color_chances)
    return np.floor(max_color * cdf).astype(int)


def by_percentiles(
    min_p: float, max_p: float, thermal: Generic2DArray[np.float64]
) -> Generic2DArray[np.float64]:
    """Normalizes the given thermal array using the values at the given minimum
    and maximum percentiles.

    Parameters
    ----------
    min_p: float
        Value between 0 and 1 that is lower than `max_p`. Indicates the
        percentile from which to take the minimum normalization value.
    max_p: float
        Value between 0 and 1 that is greater than `min_p`. Indicates the
        percentile from which to take the maximum normalization value.
    thermal: Generic2DArray[np.float64]
        Array to normalize.

    Returns
    -------
    Generic2DArray[np.float64]
        The normalized array, all np.float64 values between 0 and 1.
    """
    assert 0.0 <= min_p < max_p <= 1.0
    return by_minmax(
        np.percentile(thermal, int(min_p * 100)).item(),
        np.percentile(thermal, int(max_p * 100)).item(),
        thermal,
    )



from typing import BinaryIO, List, Tuple

import numpy as np
from PIL import Image

from flyr.dtypes import RGBArray
from flyr.palettes.embedded import ycc_to_rgb


class Palette:
    num_colors: int
    above_color: Tuple[int, int, int]
    below_color: Tuple[int, int, int]
    overflow_color: Tuple[int, int, int]
    underflow_color: Tuple[int, int, int]
    isotherm1_color: Tuple[int, int, int]
    isotherm2_color: Tuple[int, int, int]
    method: int
    stretch: int
    file_name: str
    name: str
    yccs: List[Tuple[int, int, int]]
    rgbs: List[Tuple[int, int, int]]

    def __init__(
        self,
        num_colors: int,
        above_color: Tuple[int, int, int],
        below_color: Tuple[int, int, int],
        overflow_color: Tuple[int, int, int],
        underflow_color: Tuple[int, int, int],
        isotherm1_color: Tuple[int, int, int],
        isotherm2_color: Tuple[int, int, int],
        method: int,
        stretch: int,
        file_name: str,
        name: str,
        yccs: List[Tuple[int, int, int]],
    ):
        self.num_colors = num_colors
        self.above_color = above_color
        self.below_color = below_color
        self.overflow_color = overflow_color
        self.underflow_color = underflow_color
        self.isotherm1_color = isotherm1_color
        self.isotherm2_color = isotherm2_color
        self.method = method
        self.stretch = stretch
        self.file_name = file_name
        self.name = name
        self.yccs = yccs
        self.rgbs = [ycc_to_rgb(*yccs[i]) for i in range(num_colors)]

        assert len(self.yccs) > 0
        assert len(self.yccs) == len(self.rgbs)

    def render(self) -> RGBArray:
        img = np.zeros((20, self.num_colors * 2, 3), dtype=np.uint8)
        for i, (r, g, b) in enumerate(self.rgbs):
            img[:, i * 2 : (i + 1) * 2, 0] = r
            img[:, i * 2 : (i + 1) * 2, 1] = g
            img[:, i * 2 : (i + 1) * 2, 2] = b
        return img

    def render_pil(self) -> Image.Image:
        return Image.fromarray(self.render())


def parse_palette_info(stream: BinaryIO, metadata: Tuple[int, int, int, int]) -> Palette:
    (_, _, offset, _) = metadata

    stream.seek(offset)
    num_colors = int.from_bytes(stream.read(6), "little")

    def read_color_tuple(stream):
        c1 = int.from_bytes(stream.read(1), "little")
        c2 = int.from_bytes(stream.read(1), "little")
        c3 = int.from_bytes(stream.read(1), "little")
        return (c1, c2, c3)

    above = read_color_tuple(stream)
    below = read_color_tuple(stream)
    overflow = read_color_tuple(stream)
    underflow = read_color_tuple(stream)
    isotherm1 = read_color_tuple(stream)
    isotherm2 = read_color_tuple(stream)
    stream.read(2)
    method = int.from_bytes(stream.read(1), "little")
    stretch = int.from_bytes(stream.read(1), "little")

    stream.seek(offset + 112)
    palette: bytes = stream.read(num_colors * 3)
    palette_ycc: List[Tuple[int, int, int]] = [
        (palette[i], palette[i + 1], palette[i + 2]) for i in range(0, len(palette), 3)
    ]

    return Palette(
        num_colors,
        above,
        below,
        overflow,
        underflow,
        isotherm1,
        isotherm2,
        method,
        stretch,
        "",  # TODO Read palette file name
        "",  # TODO Read palette name
        palette_ycc,
    )



from enum import Enum
from io import BytesIO
from typing import BinaryIO, Dict, Optional, Tuple


# Constants
segment_sep = b"\xff"
app1_marker = b"\xe1"
magic_flir_def = b"FLIR\x00"

chunk_app1_bytes_count = len(app1_marker)
chunk_length_bytes_count = 2
chunk_magic_bytes_count = len(magic_flir_def)
chunk_skip_bytes_count = 1
chunk_num_bytes_count = 1
chunk_tot_bytes_count = 1
chunk_partial_metadata_length = (
    chunk_app1_bytes_count + chunk_length_bytes_count + chunk_magic_bytes_count
)
chunk_metadata_length = (
    chunk_partial_metadata_length
    + chunk_skip_bytes_count
    + chunk_num_bytes_count
    + chunk_tot_bytes_count
)


class RecordIndex(Enum):
    RAW_DATA = 1
    EMBEDDED_IMAGE = 14
    CAMERA_INFO = 32
    MEASUREMENT_INFO = 33
    PALETTE_INFO = 34
    TEXT_INFO = 35
    PICTURE_IN_PICTURE_INFO = 42


def extract_flir_app1(stream: BinaryIO) -> BinaryIO:
    """Extracts the FLIR APP1 bytes.

    Parameters
    ---------
    stream: BinaryIO
        A full bytes stream of a JPEG file, expected to be a FLIR file.

    Raises
    ------
    ValueError
        When the file is invalid in one the next ways, a
        ValueError is thrown.

        * File is not a JPEG
        * A FLIR chunk number occurs more than once
        * The total chunks count is inconsistent over multiple chunks
        * No APP1 segments are successfully parsed

    Returns
    -------
    BinaryIO
        A bytes stream of the APP1 FLIR segments
    """
    # Check JPEG-ness
    magic_bytes = stream.read(2)
    if False:  # TODO check whether image is a correct jpeg
        raise ValueError("Invalid FLIR: given file is not a jpeg")

    chunks_count: Optional[int] = None
    chunks: Dict[int, bytes] = {}
    while True:
        b = stream.read(1)
        if b == b"":
            break

        if b != segment_sep:
            continue

        parsed_chunk = parse_flir_chunk(stream, chunks_count)
        if not parsed_chunk:
            continue

        chunks_count, chunk_num, chunk = parsed_chunk
        chunk_exists = chunks.get(chunk_num, None) is not None
        if chunk_exists:
            raise ValueError("Invalid FLIR: duplicate chunk number")
        chunks[chunk_num] = chunk

        # Encountered all chunks, break out of loop to process found metadata
        if chunk_num == chunks_count:
            break

    if chunks_count is None:
        raise ValueError("Invalid FLIR: no metadata encountered")

    flir_app1_bytes = b""
    for chunk_num in range(chunks_count + 1):
        flir_app1_bytes += chunks[chunk_num]

    flir_app1_stream = BytesIO(flir_app1_bytes)
    flir_app1_stream.seek(0)
    return flir_app1_stream


def parse_flir_chunk(
    stream: BinaryIO, chunks_count: Optional[int]
) -> Optional[Tuple[int, int, bytes]]:
    # Parse the chunk header. Headers are as follows (definition with example):
    #
    #     \xff\xe1<length: 2 bytes>FLIR\x00\x01<chunk nr: 1 byte><chunk count: 1 byte>
    #     \xff\xe1\xff\xfeFLIR\x00\x01\x01\x0b
    #
    # Meaning: Exif APP1, 65534 long, FLIR chunk 1 out of 12
    marker = stream.read(chunk_app1_bytes_count)

    length_bytes = stream.read(chunk_length_bytes_count)
    length = int.from_bytes(length_bytes, "big")
    length -= chunk_metadata_length
    magic_flir = stream.read(chunk_magic_bytes_count)

    if not (marker == app1_marker and magic_flir == magic_flir_def):
        # Seek back to just after byte b and continue searching for chunks
        stream.seek(-len(marker) - len(length_bytes) - len(magic_flir), 1)
        return None

    stream.seek(1, 1)  # skip 1 byte, unsure what it is for

    chunk_num = int.from_bytes(stream.read(chunk_num_bytes_count), "big")
    chunks_tot = int.from_bytes(stream.read(chunk_tot_bytes_count), "big")

    # Remember total chunks to verify metadata consistency
    if chunks_count is None:
        chunks_count = chunks_tot

    if (  # Check whether chunk metadata is consistent
        chunks_tot is None
        or chunk_num < 0
        or chunk_num > chunks_tot
        or chunks_tot != chunks_count
    ):
        raise ValueError(
            f"Invalid FLIR: inconsistent total chunks, should be 0 or greater, "
            "but is {chunks_tot}"
        )

    # FIXME There is  calculation error somewhere which causes length to be
    #       1 too small. So far it manually adding 1 seems sufficient to
    #       correct for it, but it's not the right way.
    #       Fix by figuring out what causes the off by 1 error.
    return chunks_tot, chunk_num, stream.read(length + 1)


def parse_flir_records(stream: BinaryIO) -> Dict[int, Tuple[int, int, int, int]]:
    # 0x00 - string[4] file format ID = "FFF\0"
    # 0x04 - string[16] file creator: seen "\0","MTX IR\0","CAMCTRL\0"
    # 0x14 - int32u file format version = 100
    # 0x18 - int32u offset to record directory
    # 0x1c - int32u number of entries in record directory
    # 0x20 - int32u next free index ID = 2
    # 0x24 - int16u swap pattern = 0 (?)
    # 0x28 - int16u[7] spares
    # 0x34 - int32u[2] reserved
    # 0x3c - int32u checksum

    # 1. Read 0x40 bytes and verify that its contents equals AFF\0 or FFF\0
    file_format_id = stream.read(4)  # TODO the check

    # 2. Read FLIR record directory metadata (ref 3)
    stream.seek(16, 1)
    file_format_version = int.from_bytes(stream.read(4), "big")
    record_dir_offset = int.from_bytes(stream.read(4), "big")
    record_dir_entries_count = int.from_bytes(stream.read(4), "big")
    stream.seek(28, 1)
    checksum = int.from_bytes(stream.read(4), "big")

    # 3. Read record directory (which is a FLIR record entry repeated
    # `record_dir_entries_count` times)
    stream.seek(record_dir_offset)
    record_dir_stream = BytesIO(stream.read(32 * record_dir_entries_count))

    # First parse the record metadata
    record_details: Dict[int, Tuple[int, int, int, int]] = {}
    for record_nr in range(record_dir_entries_count):
        record_dir_stream.seek(0)
        details = parse_flir_record_metadata(stream, record_nr)
        if details:
            record_details[details[1]] = details

    # Then parse the actual records
    # for (entry_idx, type, offset, length) in record_details:
    #     parse_record = record_parsers[type]
    #     stream.seek(offset)
    #     record = BytesIO(stream.read(length + 36))  # + 36 needed to find end
    #     parse_record(record, offset, length)

    return record_details


def parse_flir_record_metadata(
    stream: BinaryIO, record_nr: int
) -> Optional[Tuple[int, int, int, int]]:
    # FLIR record entry (ref 3):
    # 0x00 - int16u record type
    # 0x02 - int16u record subtype: RawData 1=BE, 2=LE, 3=PNG; 1 for other record types
    # 0x04 - int32u record version: seen 0x64,0x66,0x67,0x68,0x6f,0x104
    # 0x08 - int32u index id = 1
    # 0x0c - int32u record offset from start of FLIR data
    # 0x10 - int32u record length
    # 0x14 - int32u parent = 0 (?)
    # 0x18 - int32u object number = 0 (?)
    # 0x1c - int32u checksum: 0 for no checksum
    entry = 32 * record_nr
    stream.seek(entry)
    record_type = int.from_bytes(stream.read(2), "big")
    if record_type < 1:
        return None

    # TODO convert unnecessary reads to seeks
    stream.seek(10, 1)
    # record_subtype = int.from_bytes(stream.read(2), "big")
    # record_version = int.from_bytes(stream.read(4), "big")
    # index_id = int.from_bytes(stream.read(4), "big")
    record_offset = int.from_bytes(stream.read(4), "big")
    record_length = int.from_bytes(stream.read(4), "big")
    stream.seek(12, 1)
    # parent = int.from_bytes(stream.read(4), "big")
    # object_numer = int.from_bytes(stream.read(4), "big")
    # checksum = int.from_bytes(stream.read(4), "big")

    return (entry, record_type, record_offset, record_length)



#!/usr/bin/env python3

"""Main module containing the code that parses/unpacks a FLIR file.

The most important function here for end users is `unpack()` which starts
the unpacking sequence. It returns a `FlyrThermogram`, defined in
thermogram.py, so check there for the main way to interface with the
results.

Much of the documentation needed to implement this was gotten from
exiftool (https://exiftool.org/TagNames/FLIR.html).
"""

# FIXME stop depending on PIL and nptyping

# Standard library imports
import os
from io import BufferedIOBase
from typing import BinaryIO, Callable, Dict, List, Tuple, TypeVar, Union

# Self imports
from .camera_metadata import CameraMetadata
from .embedded_image import parse_embedded_image
from .measurement_info import Measurement, parse_measurements
from .palette_info import parse_palette_info
from .picture_in_picture import parse_pip_info
from .records import RecordIndex, extract_flir_app1, parse_flir_records
from .thermal import parse_camera_info, parse_raw_data
from .thermogram import FlyrThermogram

T = TypeVar("T")


def unpack(path_or_stream: Union[str, BinaryIO]) -> FlyrThermogram:
    """Unpacks the FLIR image, meaning that it will return the thermal data
    embedded in the image.

    Parameters
    ----------
    path_or_stream : Union[str, BinaryIO]
        Either a path (string) to a FLIR file, or a byte stream such as
        BytesIO or file opened as `open(file_path, "rb")`.

    Returns
    -------
    FlyrThermogram
        When successful, a FlyrThermogram object containing thermogram data.
    """
    if isinstance(path_or_stream, str) and os.path.isfile(path_or_stream):
        with open(path_or_stream, "rb") as flirh:
            thermogram = unpack(flirh)
        thermogram.path = path_or_stream
        return thermogram
    elif isinstance(path_or_stream, BufferedIOBase):
        stream = path_or_stream
        flir_app1_stream = extract_flir_app1(stream)
        flir_records = parse_flir_records(flir_app1_stream)
        stream.seek(0)
        thermogram = parse_thermogram(flir_app1_stream, flir_records, stream)

        return thermogram
    else:
        raise ValueError("Incorrect input")  # TODO improved error message


def parse_thermogram(
    thermal_stream: BinaryIO,
    records: Dict[int, Tuple[int, int, int, int]],
    file_stream: BinaryIO,
) -> FlyrThermogram:
    # FIXME Support formats other than subtype == PNG
    # Parse core thermal data: raw values and camera settings
    raw_data_md = records[RecordIndex.RAW_DATA.value]
    _, _, raw_data = parse_raw_data(thermal_stream, raw_data_md)

    camera_info_md = records[RecordIndex.CAMERA_INFO.value]
    camera_info = parse_camera_info(thermal_stream, camera_info_md)

    # Parse optional embedded information
    def _try(
        parse: Callable[[BinaryIO, Tuple[int, int, int, int]], T],
        index: RecordIndex,
        fallback: T,
    ) -> T:
        return _try_parse(thermal_stream, records, index, parse, fallback)

    embedded_image_ = _try(parse_embedded_image, RecordIndex.EMBEDDED_IMAGE, None)
    embedded_image = None if embedded_image_ is None else embedded_image_[2]
    palette_info = _try(parse_palette_info, RecordIndex.PALETTE_INFO, None)
    pip_info = _try(parse_pip_info, RecordIndex.PICTURE_IN_PICTURE_INFO, None)
    measurements: List[Measurement] = []
    measurements = _try(parse_measurements, RecordIndex.MEASUREMENT_INFO, measurements)

    # Finish with return object
    thermogram = FlyrThermogram(
        raw_data,
        camera_info,
        optical=embedded_image,
        palette=palette_info,
        picture_in_picture=pip_info,
        measurements=measurements,
        metadata_camera=CameraMetadata(file_stream),
    )
    return thermogram


def _try_parse(
    stream: BinaryIO,
    records: Dict[int, Tuple[int, int, int, int]],
    index: RecordIndex,
    parse: Callable[[BinaryIO, Tuple[int, int, int, int]], T],
    fallback: T,
) -> T:
    try:
        md = records[index.value]
        value = parse(stream, md)
    except:  # noqa: E722
        value = fallback
    return value


def main():
    import argparse
    import traceback as tb
    import warnings
    from pprint import pprint

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--camera-info", "-c", action="store_true", default=False, required=False
    )
    parser.add_argument("--palette-dest", default="", required=False)
    parser.add_argument("--render-dest", "-r", default="", required=False)
    parser.add_argument("--optical-dest", "-o", default="", required=False)
    parser.add_argument("--pnp-dest", "-p", default="", required=False)
    parser.add_argument("--emboss-dest", "-e", default="", required=False)
    parser.add_argument(
        "--print-measurements", "-m", action="store_true", default=False, required=False
    )
    parser.add_argument(
        "--correct-msx", "--msx", action="store_true", default=False, required=False
    )
    parser.add_argument(
        "--equalize-histogram",
        "--eh",
        action="store_true",
        default=False,
        required=False,
    )
    parser.add_argument("thermogram")
    args = parser.parse_args()

    script = os.path.basename(__file__)
    if not os.path.isfile(args.thermogram):
        print(f"{script}: error: file {args.thermogram} doesn't exist")
        exit(1)

    try:
        thermogram = unpack(args.thermogram)

        if args.camera_info:
            pprint(thermogram.metadata)

        if args.palette_dest and thermogram.palette is not None:
            thermogram.palette.render_pil().save(args.palette_dest)
        elif args.palette_dest:
            warnings.warn(f"{script}: warning: no palette was extracted")

        if args.render_dest:
            render = thermogram.render_pil(histogram_equalized=args.equalize_histogram)
            render.save(args.render_dest)

        if args.optical_dest and thermogram.optical_pil is not None:
            thermogram.optical_pil.save(args.optical_dest)

        if args.pnp_dest:
            msx_correction = args.correct_msx
            render = thermogram.picture_in_picture_pil(msx_correction=msx_correction)
            render.save(args.pnp_dest)

        if args.emboss_dest:
            msx_correction = args.correct_msx
            render = thermogram.render_pil(
                edge_emphasis=0.275, msx_correction=msx_correction
            )
            render.save(args.emboss_dest)

        if args.print_measurements:
            for m in thermogram.measurements:
                ident = f"{m.tool.name.capitalize()} {m.label}".strip()
                print(f"{ident}: {m.params}")
    except Exception as e:
        msg = f"{tb.format_exc()}\n{script}: error: failed unpacking {args.thermogram}"
        print(msg)


if __name__ == "__main__":
    main()



"""Module containing the implementation of image alignment algorithms."""

import numpy as np

from flyr.dtypes import Generic2DArray, MaskArray

__EPS: float = 1e-9  # Division by 0 prevention


def calculate_translation(
    image_1: MaskArray, image_2: MaskArray, threshold: int = 8
) -> tuple[int, int]:
    """Calculate the translation between two images using phase correlation. Return
    (dy, dx) translation that aligns `image_2` to `image_1`.

    Parameters
    ----------
    image_1 : MaskArray
        The first image.
    image_2 : MaskArray
        The second image.
    threshold : int, optional
        The minimum Peak to Second Peak Ratio to consider the translation valid, by default 8.

    Returns
    -------
    tuple[int, int]
        The translation between the two images.
    """
    if image_1.shape != image_2.shape:
        msg = "To estimate phase correlation, the inputs must have the same shape"
        raise ValueError(msg)

    matrix = _calculate_phase_correlation_matrix(image_1, image_2)
    matrix = _filter_non_maximum_peaks(matrix)
    max_y, max_x = _find_maximum_peaks_coords(matrix)

    metric = _calculate_metric(matrix, max_x, max_y)
    if metric > threshold:
        h, w = image_1.shape
        dy = max_y if max_y < h // 2 else max_y - h
        dx = max_x if max_x < w // 2 else max_x - w
        return int(dx), int(dy)
    return 0, 0


def _calculate_metric(
    matrix: Generic2DArray, max_x: int, max_y: int, window: int = 5
) -> float:
    """Calculate the Peak to Second Peak Ratio to understand if the maximum peak is stable.

    Parameters
    ----------
    max_x : int
        The x coordinate of the maximum peak.
    max_y : int
        The y coordinate of the maximum peak.
    window : int, optional
        The size of the window to exclude around the maximum peak, by default 5.

    Returns
    -------
    float
        The Peak to Second Peak Ratio.
    """

    h, w = matrix.shape
    peak = matrix[max_y, max_x]

    mask = np.ones_like(matrix, dtype=bool)
    y0 = max(0, max_y - window)
    y1 = min(h, max_y + window + 1)
    x0 = max(0, max_x - window)
    x1 = min(w, max_x + window + 1)

    mask[y0:y1, x0:x1] = False

    sidelobes = matrix[mask]
    mean = sidelobes.mean()
    std = sidelobes.std()

    return (peak - mean) / (std + __EPS)


def _calculate_phase_correlation_matrix(
    image_1: MaskArray, image_2: MaskArray
) -> Generic2DArray:
    """Calculate the phase correlation matrix between two images.

    Parameters
    ----------
    image_1 : MaskArray
        The first image.
    image_2 : MaskArray
        The second image.

    Returns
    -------
    Generic2DArray
        The phase correlation matrix.
    """
    f_image_1 = np.fft.fft2(image_1)
    f_image_2 = np.fft.fft2(image_2)

    r = f_image_1 * np.conj(f_image_2)
    r /= np.maximum(np.abs(r), __EPS)
    return np.abs(np.fft.ifft2(r))


def _filter_non_maximum_peaks(matrix: Generic2DArray) -> Generic2DArray:
    """Subtract the matrix by estimating the background as the median to filter the non-maximum peaks.

    Parameters
    ----------
    matrix : Generic2DArray
        The matrix to filter.

    Returns
    -------
    Generic2DArray
        The filtered matrix.
    """
    matrix -= np.median(matrix)
    return np.clip(matrix, 0, None)


def _find_maximum_peaks_coords(matrix: Generic2DArray) -> tuple[int, int]:
    """Find the coordinates of the maximum peak in the matrix.

    Returns
    -------
    Generic2DArray
        The coordinates of the maximum peak.
    """
    indices = np.unravel_index(np.argmax(matrix), matrix.shape)
    return (int(indices[0]), int(indices[1]))



from io import BytesIO
from typing import BinaryIO, Tuple
from flyr.dtypes import GrayArray

import numpy as np
from PIL import Image


def parse_embedded_image(
    stream: BinaryIO, metadata: Tuple[int, int, int, int]
) -> Tuple[int, int, GrayArray]:
    (entry_idx, _, offset, length) = metadata
    stream.seek(offset)

    # from PIL import ImageFile  # Uncomment to enable loading corrupt PNGs
    # ImageFile.LOAD_TRUNCATED_IMAGES = True

    stream.seek(2, 1)  # Skip first two bytes, TODO Explain the why of the skip
    width = int.from_bytes(stream.read(2), "little")
    height = int.from_bytes(stream.read(2), "little")

    stream.seek(offset + 2 * 16)  # TODO document why 2 * 16

    # Read the bytes with the embedded image and decode using PIL
    optical_bytes = stream.read(length)
    optical_stream = BytesIO(optical_bytes)
    optical_img = Image.open(optical_stream)
    optical_np = np.array(optical_img)

    # Check shape
    if optical_np.shape[:2] != (height, width):
        msg = "Invalid FLIR: metadata's width and height don't match optical data's actual width and height ({} vs ({}, {})"
        msg = msg.format(optical_np.shape, height, width)
        raise ValueError(msg)

    return width, height, optical_np



import struct
from typing import BinaryIO, Tuple


class PictureInPicture:
    real_to_ir: float
    offset_x: int
    offset_y: int
    pip_x1: int
    pip_x2: int
    pip_y1: int
    pip_y2: int

    def __init__(
        self,
        real_to_ir: float,
        offset_x: int,
        offset_y: int,
        pip_x1: int,
        pip_x2: int,
        pip_y1: int,
        pip_y2: int,
    ):
        self.real_to_ir = real_to_ir
        self.offset_x = offset_x
        self.offset_y = offset_y
        self.pip_x1 = pip_x1
        self.pip_x2 = pip_x2
        self.pip_y1 = pip_y1
        self.pip_y2 = pip_y2

    @property
    def crop_box(self) -> Tuple[int, int, int, int]:
        return (self.pip_x1, self.pip_y1, self.pip_x2, self.pip_y2)


def parse_picture_in_picture_info(
    stream: BinaryIO, metadata: Tuple[int, int, int, int]
) -> PictureInPicture:
    (_, _, offset, _) = metadata

    stream.seek(offset)

    ratio = stream.read(4)
    real_to_ir = struct.unpack("f", ratio)[0]
    offset_x = int.from_bytes(stream.read(2), "little", signed=True)
    offset_y = int.from_bytes(stream.read(2), "little", signed=True)
    pip_x1 = int.from_bytes(stream.read(2), "little", signed=True)
    pip_x2 = int.from_bytes(stream.read(2), "little", signed=True)
    pip_y1 = int.from_bytes(stream.read(2), "little", signed=True)
    pip_y2 = int.from_bytes(stream.read(2), "little", signed=True)

    return PictureInPicture(
        real_to_ir, offset_x, offset_y, pip_x1, pip_x2, pip_y1, pip_y2
    )


def parse_pip_info(
    stream: BinaryIO, metadata: Tuple[int, int, int, int]
) -> PictureInPicture:
    return parse_picture_in_picture_info(stream, metadata)



from typing import Annotated, Literal, TypeVar, Union

import numpy as np
import numpy.typing as npt

DType = TypeVar("DType", bound=np.generic)

# Masks are type bool and have 2 dimensions
MaskArray = Annotated[npt.NDArray[np.bool_], Literal[2]]

# Grayscale images are type uint8 and have 2 dimensions
GrayArray = Annotated[npt.NDArray[np.uint8], Literal[2]]

# RGB images are type uint8 and have 3 dimensions
RGBArray = Annotated[npt.NDArray[np.uint8], Literal[3]]

# Generic 1D, 2D and 3D arrays
Generic1DArray = Annotated[npt.NDArray[DType], Literal[1]]
Generic2DArray = Annotated[npt.NDArray[DType], Literal[2]]
Generic3DArray = Annotated[npt.NDArray[DType], Literal[3]]

# Raw value arrays are type uint16 (raw bytes) or int64 (PNG) and have 2 dimensions
RawArray = Union[Generic2DArray[np.int64], Generic2DArray[np.uint16]]

Temperature = Literal["kelvin", "celsius", "fahrenheit"]



"""The main public API to the data in a FLIR thermogram and the result of
`flyr.unpack()`.
"""

import os
import warnings
from math import exp, fsum, sqrt
from typing import Dict, List, Literal, Optional, Tuple, Union

import numpy as np
from PIL import Image, ImageFilter

import flyr.camera_metadata as cm
import flyr.edges as edg
import flyr.measurement_info as mi
import flyr.msx as msx
import flyr.normalization as norm
import flyr.palette_info as pal
import flyr.palettes as palettes
import flyr.picture_in_picture as pip
from flyr.dtypes import (
    Generic2DArray,
    MaskArray,
    RawArray,
    RGBArray,
    Temperature,
)


class FlyrThermogram:
    """A FlyrThermogram is a class providing read-only access to the data in a typical
    FLIR thermogram.

    Specifically interesting are:

    * `kelvin` (property): Getting the temperature in degrees kelvin
    * `celsius` (property): Getting the temperature in degrees celsius
    * `metadata` (property): Getting the FLIR camera metadata, some of which
        influences how `kelvin`/`celsius` are calculated.
    * `adjust_metadata()` (method): Updates the above-mentioned metadata and can thus
        be used to change how the temperature is calculated.
    * `render()` (method): Returns the temperature as an RGB array. Use the
        `render_pil` variant to get one
    """

    # Required members variables required to be set
    __metadata_adjustments: Dict[str, Union[float, int]]
    __metadata: Dict[str, Union[float, int]]
    __optical: Optional[RGBArray]
    __thermal: RawArray

    camera_metadata: Optional[cm.CameraMetadata]
    measurements: List[mi.Measurement]
    palette: Optional[pal.Palette]
    path: Optional[str]
    pip_info: Optional[pip.PictureInPicture]

    def __init__(
        self,
        thermal: RawArray,
        metadata: Dict[str, Union[float, int]],
        optical: Optional[RGBArray] = None,
        palette: Optional[pal.Palette] = None,
        picture_in_picture: Optional[pip.PictureInPicture] = None,
        path: Optional[str] = None,
        metadata_adjustments: Dict[str, Union[float, int]] = {},
        metadata_camera: Optional[cm.CameraMetadata] = None,
        measurements: List[mi.Measurement] = [],
    ):
        """Initialize a new instance of this class. The raw thermal data and
        the accompanying metadata to correctly interpret this data are
        required.

        Parameters
        ----------
        thermal: RawArray
            A 2D numpy array of 64 or 16 bit integers. This is the raw thermal
            data as it is stored in the FLIR file. Order is [H, W].
        metadata: Dict[str, Union[float, int]]
            A dictionary with physical parameters to interpret the raw
            thermal data, as it was included in the original FLIR file.
        optical: RGBArray
            A 3D numpy array of 8 bit integers, in the order of [H, W, C].
            This should be a 'normal' photo (RGB) of the same scene as
            thermogram.
        palette: Optional[pinfo.Palette]
            The palette embedded in the file, which can be used to render
            with by default. Is `None` by default in which case a grayscale
            palette is used.
        picture_in_picture: Optional[pip.PictureInPicture]
            The picture-in-picture settings useful for stacking optical and
            thermal data.
        path: Optional[str]
            The path to the original file. Is only used to determine this
            this object's identifier and can safely be left `None`
            (default).
        metadata_adjustments: Dict[str, Union[float, int]]
            A dictionary with adjustments to the above dictionary. This
            separation allows both using the original settings and
            calculating temperatures with different physical parameters.
        metadata_camera: Optional[cm.CameraMetadata]
        measurements: List[mi.Measurement]

        Returns
        -------
        FlyrThermogram
        """
        self.__metadata = metadata.copy()
        self.__metadata_adjustments = metadata_adjustments.copy()
        self.__optical = optical  # Optical (RGB) photo in the thermogram
        self.__thermal = thermal  # Raw thermal data
        self.camera_metadata = metadata_camera
        self.measurements = measurements
        self.palette = palette
        self.path = path
        self.pip_info = picture_in_picture

    @property
    def kelvin(self) -> Generic2DArray[np.float64]:
        """A property method that returns the thermogram's temperature in
        kelvin (K).

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in kelvin. Order is [H, W].
        """
        return self.__raw_to_kelvin_with_metadata(self.__thermal)

    @property
    def celsius(self) -> Generic2DArray[np.float64]:
        """Returns the thermogram's temperature in degrees celsius (°C).

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in celsius. Order is [H, W].
        """
        return self.kelvin - 273.15

    @property
    def fahrenheit(self) -> Generic2DArray[np.float64]:
        """Returns the thermogram's temperature in degrees fahrenheit (°F).

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in fahrenheit. Order is [H, W].
        """
        return self.celsius * 1.8 + 32.00

    @property
    def reaumur(self) -> Generic2DArray[np.float64]:
        """Returns the thermogram's temperature in degrees reaumur (°Re), a
        historical scale preceding Celsius proposed in 1730.

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in fahrenheit. Order is [H, W].
        """
        return self.celsius * 0.8

    @property
    def rankine(self) -> Generic2DArray[np.float64]:
        """Returns the thermogram's temperature in degrees rankine (°Ra), a
        historical scale from 1859 similar to Fahrenheit.

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in fahrenheit. Order is [H, W].
        """
        return self.kelvin * 1.8

    @property
    def delisle(self) -> Generic2DArray[np.float64]:
        """Returns the thermogram's temperature in degrees rankine (°D), a
        historical scale from 1732.

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in fahrenheit. Order is [H, W].
        """
        return (375.15 - self.kelvin) * 1.5

    @property
    def newton(self) -> Generic2DArray[np.float64]:
        """Returns the thermogram's temperature in degrees Newton (°N), a
        historical scale devised in 1701.

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in fahrenheit. Order is [H, W].
        """
        return self.celsius * (33 / 100)

    @property
    def roemer(self) -> Generic2DArray[np.float64]:
        """Returns the thermogram's temperature in degrees Rømer (°Rø), a
        historical scale from 1702.

        Returns
        -------
        Generic2DArray[np.float64]
            A 2D array of numpy float values in fahrenheit. Order is [H, W].
        """
        return self.celsius * (21 / 40) + 7.5

    @property
    def optical(self) -> Optional[RGBArray]:
        """Returns the thermogram's embedded photo.

        Returns
        -------
        RGBArray or `None`
            A 3D array of 8 bit integers containing the RGB photo
            embedded within the FLIR thermogram.  Order is [H, W, C].
        """
        return None if self.__optical is None else self.__optical.copy()

    @property
    def optical_pil(self) -> Optional[Image.Image]:
        """Returns the thermogram's embedded photo as a Pillow `Image`.

        Returns
        -------
        `PIL.Image`
            A Pillow Image object of the RGB photo embedded within the FLIR
            thermogram.
        """
        optical = self.optical
        return None if optical is None else Image.fromarray(optical)

    @property
    def metadata(self) -> Dict[str, Union[float, int]]:
        metadata = self.__metadata.copy()
        metadata.update(self.__metadata_adjustments)
        return metadata

    @property
    def identifier(self) -> Optional[str]:
        return os.path.basename(self.path) if self.path is not None else None

    def render(
        self,
        min_v: Optional[float] = None,
        max_v: Optional[float] = None,
        unit: Temperature = "kelvin",
        palette: Union[palettes.Palette, palettes.PaletteColors] = "embedded",
        mode: norm.NormalizationMode = "minmax",
        histogram_equalized: bool = False,
    ) -> RGBArray:
        """Renders the thermogram to RGB with the given settings.

        First the thermogram is normalized using the given interval and
        mode. Then the palette is used to translate the values to colors.

        Parameters
        ----------
        min_v: float or `None`. Is `None` by default.
            When set to `None`, the lower bound the FLIR camera used to render the
            original file with is used.
                All values below this value will be clipped to this value,
            although the exact behaviour depends on the `unit`.
                When unit is celsius, kelvin or fahrenheit, the `min_v` and `max_v`
            values function directly as the thresholds to which the thermogram is
            clipped.
                When `unit='percentiles'`, then `min_v` and `max_v` are interpreted
            as percentiles. First the values for those percentiles are retrieved
            which are then used to clip the thermogram as described when
            `unit='kelvin'`.
        max_v: float or `None`. Is `None` by default.
            See the `min_v` for details on how it is interpreted.
        unit: Temperature
            The unit of the `min_v` and `max_v` parameters, which can be celsius,
            fahrenheit or kelvin. Default is 'kelvin'. Only used when `mode` (see
            below) is 'minmax', thus ignored in the case of 'percentiles'.
        palette: str. Is `"embedded"` by default.
            The name of the color palette to use. See the `palettes` module to see
            which are supported in addition to "embedded".
                Alternatively, a list of 3-tuples can be passed in. Each
            3-tuple has 3 integers corresponding to the RGB components.
        mode: str, one of "minmax" | "percentiles". "minmax" is default.
            How to pre-process the data before rendering. Minmax is simply clips
            the data by a minimal and maximum value; percentiles does so
            by extracting a certain min/max percentile.
        histogram_equalized: bool
            Default is `False`. Enable this to enhance contrast and better see some
            objects. Be aware that this changes thermal pixels' apparant temperature,
            so only use for presentation purposes, not analytical ones. Implemented
            per the white paper 'Histogram equalization', but achieves slightly
            different results.

        Returns
        -------
        RGBArray
            A three dimensional array of integers between 0 and 255,
            representing an RGB render of the thermogram. Order is
            [H, W, C].
        """
        normalizer = {  # Functions defined below
            "minmax": norm.by_minmax,
            "percentiles": norm.by_percentiles,
        }

        # Validate parameters
        assert min_v is None or max_v is None or min_v < max_v
        assert not ((min_v is None or max_v is None) and unit == "percentiles")
        assert unit in ["kelvin", "celsius", "fahrenheit", "percentiles"]
        assert (
            palette in ["grayscale", "grayscale-inverted", "embedded"]
            or palette in palettes.palettes
        )
        assert mode in normalizer.keys()

        # In case the min / max values are None, find the right defaults
        emb_min_v, emb_max_v = self.embedded_range(unit)
        min_v = emb_min_v if min_v is None else min_v
        max_v = emb_max_v if max_v is None else max_v

        # Convert min/max values to kelvin if necessary
        if unit == "celsius":
            min_v = 273.15 + min_v
            max_v = 273.15 + max_v
        elif unit == "fahrenheit":
            min_v = 273.15 + (min_v - 32.0) / 1.8
            max_v = 273.15 + (max_v - 32.0) / 1.8

        # Render
        normalized = normalizer[mode](min_v, max_v, self.kelvin)
        if palette == "grayscale" or palette == "grayscale-inverted":
            # Strategy for grayscale is very different from when using a map
            num_grays = 256
            if histogram_equalized:
                normalized = norm.equalize_histogram(normalized, num_grays)

            rendered = (normalized * (num_grays - 1)).astype(np.uint8)
            outshape = rendered.shape + (3,)
            repeated = np.broadcast_to(rendered[..., None], outshape)
            repeated = np.clip(repeated, 0, (num_grays - 1))
            if palette == "grayscale-inverted":
                repeated = np.invert(repeated)
            return repeated
        elif palette == "embedded":
            if self.palette is not None:
                palette = self.palette.rgbs
            else:
                palette = "turbo"
                warnings.warn(f"No embedded palette detected, using {palette} instead")
        else:
            palette = palette

        return palettes.map_colors(normalized, palette, histogram_equalized)

    def render_pil(
        self,
        edge_emphasis: float = 0.0,
        color_mask: Optional[MaskArray] = None,
        msx_correction: bool = False,
        **kwargs,
    ) -> Image.Image:
        """Renders the thermogram, but returns a Pillow Image object.

        Via Pillow, this method can offer additional functionality,
        such edge edmphasis and MSX correction.

        See `render()` for documentation on the parameters and other details.
        The parameters listed below are exclusive for `render_pil()`.

        Parameters
        ----------
        edge_emphasis: float
            Default is `0.0`. The strength with which to enhance the render's
            edges, using the optical image as a base. The value must be between
            `0.0` and `1.0`, where the first means no emphasis. No edge emphasis
            happens in case there is no optical image present.
        color_mask: Optional[Array[bool, ..., ...]]
            Default is `None`. The region where the chosen palette will be applied
            and the edges enhanced. If `None` is given, the whole thermogram will
            be colored. The mask must be of type boolean, where `True` refers
            to the region of interest. It should be same shape as the thermogram.
        msx_correction: bool
            Default is `False`. Whether to apply MSX correction to images, which
            corrects alignment between the thermal and optical images when
            overlaying them. Disabling this only uses the metadata registered in
            the image to overlay the two; enabling additionally calculates
            offsets to overlay them better.

        Raises
        ------
        ValueError
            (1) if `color_mask` is not per the requirements; (2) if
            `msx_correction` is enabled and no optical image or
            picture-in-picture data is present. (2) if `edge_emphasis` is
            enabled in the same cases.

        Returns
        -------
        PIL.Image
            A pillow Image of the rendered thermogram.
        """
        valid_color_mask = color_mask is None or (
            isinstance(color_mask, np.ndarray)
            and color_mask.dtype == "bool"
            and color_mask.shape == self.kelvin.shape
        )
        if not valid_color_mask:
            msg = "Invalid color mask: It's either not an numpy array, not of type bool or the wrong shape"
            raise ValueError(msg)

        # If a color mask is given, render a grayscale background and put the
        # configured render on top of it
        render = self.render(**kwargs)
        bg_render_pil = fg_render_pil = Image.fromarray(render)
        if color_mask is not None:
            kwargs["palette"] = "grayscale"
            bg_render_pil = Image.fromarray(self.render(**kwargs))
            roi_mask = Image.fromarray(color_mask).convert("L")
            roi_mask = roi_mask.filter(ImageFilter.BoxBlur(2))
            bg_render_pil.paste(fg_render_pil, (0, 0), roi_mask)

        # If msx correction is enabled, calculate the shift to be applied
        shift_x, shift_y = 0, 0
        if msx_correction:
            if self.__optical is None:
                msg = "Can not calculate shift between thermal and optical image: No optical image present"
                raise ValueError(msg)
            if self.pip_info is None:
                msg = "Can not calculate shift between thermal and optical image: No picture-in-picture metadata present"
                raise ValueError(msg)
            shift_x, shift_y = msx.estimate_shift(
                render[:, :, 0], self.__optical, self.pip_info
            )

        # Render edges if requested
        if edge_emphasis > 0.0:
            if self.__optical is None:
                msg = "Can not determine edges: No optical image present"
                raise ValueError(msg)
            if self.pip_info is None:
                msg = "Can not determine edges: No picture-in-picture metadata present"
                raise ValueError(msg)

            bg_render_pil = edg.emphasize_edges(
                bg_render_pil,
                self.__optical,
                self.pip_info,
                edge_emphasis,
                color_mask,
                shift_x,
                shift_y,
            )

        bg_render_pil = bg_render_pil.convert("RGB")
        bg_render_pil.info["dy"] = shift_y
        bg_render_pil.info["dx"] = shift_x
        return bg_render_pil

    def picture_in_picture_pil(
        self,
        render: Optional[Image.Image] = None,
        render_opacity: float = 1.0,
        render_crop: bool = True,
        mask_mode: Literal["exclude", "grayscale"] = "exclude",
        edge_emphasis: float = 0.0,
        thermal_mask: Optional[MaskArray] = None,
        msx_correction: bool = False,
    ) -> Image.Image:
        """Render the thermal data on top of the optical data.

        See `msx_correction` and `edge_emphasis` for interesting ways to improve
        the render.

        Parameters
        ----------
        render: Pillow Image or None
            `None` by default, rendering from scratch if so. Otherwise places
            the given render on top of the RGB photo.
        render_opacity: float
            Default 1.0. Lower this value to make the thermal data (partially)
            transparant.
        render_crop: True
            Whether or not to crop the thermal data as specified in the
            Picture-in-Picture metadata. You probably want this.
        mask_mode: str, either "exclude" or "grayscale".
            Requires `thermal_mask` to be set. `"classical"` mask mode makes all
            the thermal pixels excluded by the mask transparent. "grayscale" does
            include these thermal pixels, but renders them grayscaled.
        edge_emphases: float
            0.0 by default, effectively disabling it. Raise it to see edges
            extracted from the optical data in the final render. Requires an RGB
            photo to have been captured alongside the thermal data. Range is 0 to 1.
            A good value is 0.275.
        thermal_mask: MaskArray or None,
            Defaults to `None`. If specified, only these thermal pixels are
            rendered on top of the RGB pixels. Also see the `mask_mode` parameter.
        msx_correction: bool
            Defaults to False. Whether or not to do additionally correction of
            the placement of thermogram on top of the optical data. Also affects
            edge emphasis' placement.

        Returns
        -------
        PIL.Image.Image
            The thermal data rendered on top of the optical data.
        """
        optical_pil = self.optical_pil
        pip_info = self.pip_info
        if optical_pil is None or pip_info is None:
            raise ValueError("Impossible operation due missing information")

        assert mask_mode in ["exclude", "grayscale"]

        ratio = pip_info.real_to_ir
        offset_x = pip_info.offset_x
        offset_y = pip_info.offset_y
        render_opacity = min(max(render_opacity, 0.0), 1.0)

        # Get the optical and render
        optical_pil = optical_pil.convert("RGBA")
        if render is None:
            # Render thermogram if none given
            render_pil = self.render_pil(
                edge_emphasis=edge_emphasis,
                color_mask=thermal_mask,
                msx_correction=msx_correction,
            )

            # Ensure there is an alpha channel and make transparent if classical mask mode
            render_arr = np.array(render_pil.convert("RGBA"))
            render_arr[:, :, 3] = round(255 * render_opacity)
            if (
                thermal_mask is not None
                and mask_mode == "exclude"
                and np.any(np.invert(thermal_mask))
            ):
                render_arr[np.invert(thermal_mask)] = 0

            render_pil = Image.fromarray(render_arr)
        else:
            render_pil = render.convert("RGBA")

        shift_x, shift_y = 0, 0
        if msx_correction:
            _optical = np.array(optical_pil)
            _render = np.array(render_pil)
            shift_x, shift_y = msx.estimate_shift(_render[:, :, 0], _optical, pip_info)

        # Resize the images according the picture-in-picture metadata
        ratio = optical_pil.size[0] / render_pil.size[0] / ratio
        if render_crop:
            render_pil = render_pil.crop(pip_info.crop_box)

        dst = (round(render_pil.size[0] * ratio), round(render_pil.size[1] * ratio))
        render_pil = render_pil.resize(dst)

        # Calculate placement and finish
        width, height = optical_pil.size
        origin_x, origin_y = width // 2, height // 2
        origin_x = origin_x - render_pil.size[0] // 2 + offset_x + shift_x
        origin_y = origin_y - render_pil.size[1] // 2 + offset_y + shift_y

        optical_pil.paste(render_pil, (origin_x, origin_y), render_pil)
        return optical_pil.convert("RGB")

    def adjust_metadata(
        self, in_place=False, **kwargs: Union[float, int]
    ) -> "FlyrThermogram":
        """Updates the physical metadata used to calculate the kelvin /
        celsius values based on the raw thermal data.

        This can be used to calculate kelvin/celsius values with different
        settings than the ones embedded in the thermogram itself during
        capture.

        This method does not check the given parameters. Wrong parameters
        names or values will be accepted without exceptions being raised.
        These exceptions will only occur when `kelvin` or `celsius` is
        accessed.

        Important: This does *not* adjust the metadata in the file itself;
        only the in-memory metadata used to calculate the temperatures
        returned by `kelvin` and `celsius` is updated. In other words, this
        method can *not* be used to modify or create a FLIR thermogram file
        with different camera settings.

        # Parameters
        in_place: boolean
            When False, a new `FlyrThermogram` object is returned after calling this
            method. When True, this object instance is modified in place and the
            object itself is returned.
        emissivity: float
        object_distance: float
        atmospheric_temperature: float
        ir_window_temperature: float
        ir_window_transmission: float
        reflected_apparent_temperature: float
        relative_humidity: float
        planck_r1: float
        planck_r2: float
        planck_b: float
        planck_f: int
        planck_o: int
        atmospheric_trans_alpha1: float
        atmospheric_trans_alpha2: float
        atmospheric_trans_beta1: float
        atmospheric_trans_beta2: float
        atmospheric_trans_x: float

        # Return
        FlyrThermogram
            When `in_place` is False, a new FlyrThermogram object with the updates
            settings. When `in_place` is True, the FlyrThermogram object on which this
            method is called.
        """
        msg = f"Parameter in_place incorrectly not of type bool but {type(in_place)}. Be sure to pass it first."
        assert isinstance(in_place, bool), msg
        if in_place:
            self.__metadata_adjustments.update(kwargs)
            return self

        metadata_adjustments = self.__metadata_adjustments.copy()
        metadata_adjustments.update(kwargs)

        return FlyrThermogram(
            self.__thermal.copy(),
            self.__metadata.copy(),
            None if self.__optical is None else self.__optical.copy(),
            palette=self.palette,
            picture_in_picture=self.pip_info,
            metadata_adjustments=metadata_adjustments,
            path=self.path,
        )

    def embedded_range(self, unit: Temperature) -> Tuple[float, float]:
        range_median = self.__metadata["raw_value_median"]
        range_half = self.__metadata["raw_value_range"] / 2

        raw_vals = np.array([[range_median - range_half, range_median + range_half]])
        min_v, max_v = self.__raw_to_kelvin_with_metadata(raw_vals, orig=True).squeeze()
        if unit == "celsius":
            min_v = min_v - 273.15
            max_v = max_v - 273.15
        elif unit == "fahrenheit":
            min_v = (min_v - 273.15) * 1.8 + 32.00
            max_v = (max_v - 273.15) * 1.8 + 32.00

        return (min_v, max_v)

    def __raw_to_kelvin_with_metadata(
        self, thermal, orig: bool = False
    ) -> Generic2DArray[np.float64]:
        metadata = self.__metadata.copy()
        if not orig:
            metadata.update(self.__metadata_adjustments)

        return FlyrThermogram.__raw_to_kelvin(
            thermal,
            metadata["emissivity"],
            metadata["object_distance"],
            metadata["atmospheric_temperature"],
            metadata["ir_window_temperature"],
            metadata["ir_window_transmission"],
            metadata["reflected_apparent_temperature"],
            metadata["relative_humidity"],
            metadata["planck_r1"],
            metadata["planck_r2"],
            metadata["planck_b"],
            metadata["planck_f"],
            metadata["planck_o"],
            metadata["atmospheric_trans_alpha1"],
            metadata["atmospheric_trans_alpha2"],
            metadata["atmospheric_trans_beta1"],
            metadata["atmospheric_trans_beta2"],
            metadata["atmospheric_trans_x"],
        )

    @staticmethod
    def __raw_to_kelvin(
        thermal,
        emissivity,
        object_distance,
        atmospheric_temperature,
        ir_window_temperature,
        ir_window_transmission,
        reflected_apparent_temperature,
        relative_humidity,
        planck_r1,
        planck_r2,
        planck_b,
        planck_f,
        planck_o,
        atmospheric_trans_alpha1,
        atmospheric_trans_alpha2,
        atmospheric_trans_beta1,
        atmospheric_trans_beta2,
        atmospheric_trans_x,
    ) -> Generic2DArray[np.float64]:
        """Use the details camera info metadata to translate the raw
        temperatures to °Kelvin.

        Parameters
        ----------
        thermal: RawArray
            The thermal data to convert from raw values to kelvin
        emissivity : float
        object_distance : float
            Unit is meters
        atmospheric_temperature : float
            Unit is Kelvin
        ir_window_temperature : float
            Unit is Kelvin
        ir_window_transmission : float
            Unit is Kelvin
        reflected_apparent_temperature : float
            Unit is Kelvin
        relative_humidity : float
            Value in 0 and 1
        planck_r1 : float
        planck_r2 : float
        planck_b : float
        planck_f : float
        planck_o : int
        atmospheric_trans_alpha1 : float
        atmospheric_trans_alpha2 : float
        atmospheric_trans_beta1 : float
        atmospheric_trans_beta2 : float
        atmospheric_trans_x : float

        Returns
        -------
        Generic2DArray[np.float64]
            An array of float64 values in kelvin.
        """
        # Transmission through window (calibrated)
        emiss_wind = 1 - ir_window_transmission
        refl_wind = 0

        # Transmission through the air
        water = relative_humidity * exp(
            1.5587
            + 0.06939 * (atmospheric_temperature - 273.15)
            - 0.00027816 * (atmospheric_temperature - 273.17) ** 2
            + 0.00000068455 * (atmospheric_temperature - 273.15) ** 3
        )

        def calc_atmos(alpha, beta):
            term1 = -sqrt(object_distance / 2)
            term2 = alpha + beta * sqrt(water)
            return exp(term1 * term2)

        atmos1 = calc_atmos(atmospheric_trans_alpha1, atmospheric_trans_beta1)
        atmos2 = calc_atmos(atmospheric_trans_alpha2, atmospheric_trans_beta2)
        tau1 = atmospheric_trans_x * atmos1 + (1 - atmospheric_trans_x) * atmos2
        tau2 = atmospheric_trans_x * atmos1 + (1 - atmospheric_trans_x) * atmos2

        # Radiance from the environment
        def plancked(t):
            planck_tmp = planck_r2 * (exp(planck_b / t) - planck_f)
            return planck_r1 / planck_tmp - planck_o

        raw_refl1 = plancked(reflected_apparent_temperature)
        raw_refl1_attn = (1 - emissivity) / emissivity * raw_refl1

        raw_atm1 = plancked(atmospheric_temperature)
        raw_atm1_attn = (1 - tau1) / emissivity / tau1 * raw_atm1

        term3 = emissivity * tau1 * ir_window_transmission
        raw_wind = plancked(ir_window_temperature)
        raw_wind_attn = emiss_wind / term3 * raw_wind

        raw_refl2 = plancked(reflected_apparent_temperature)
        raw_refl2_attn = refl_wind / term3 * raw_refl2

        raw_atm2 = plancked(atmospheric_temperature)
        raw_atm2_attn = (1 - tau2) / term3 / tau2 * raw_atm2

        subtraction = fsum(
            [
                raw_atm1_attn,
                raw_atm2_attn,
                raw_wind_attn,
                raw_refl1_attn,
                raw_refl2_attn,
            ]
        )

        raw_obj = thermal.astype(np.float64)
        raw_obj /= emissivity * tau1 * ir_window_transmission * tau2
        raw_obj -= subtraction

        # Temperature from radiance
        raw_obj += planck_o
        raw_obj *= planck_r2
        planck_term = planck_r1 / raw_obj + planck_f
        return planck_b / np.log(planck_term)



import datetime as dt
import re
from typing import Any, BinaryIO, Dict, Optional, Union

import PIL.ExifTags as pilx
import PIL.Image as pili

_search_pattern = re.compile("(.)([A-Z][a-z]+)")
_replace_pattern = re.compile("([a-z0-9])([A-Z])")


def _camel_to_snake(name):
    name = re.sub(_search_pattern, r"\1_\2", name)
    return re.sub(_replace_pattern, r"\1_\2", name).lower()


def _convert_to_degrees(value):
    d = value[0]
    m = value[1]
    s = value[2]
    return d + (m / 60.0) + (s / 3600.0)


class CameraMetadata:
    """An object to access EXIF values in the photograph (not the thermogram).
    These are mostly the same, but can differ when it comes to image width
    and height.
    """

    __exif: pili.Exif

    def __init__(self, src: Union[str, BinaryIO]):
        """Instantiate an obejct to access exif values in a human readable way.

        Parameters
        ----------
        src: str
            The path to the FLIR image to read.
        """
        with pili.open(src) as img:
            self.__exif = img.getexif()

    def __getitem__(self, key: str) -> Optional[Any]:
        data = self.data
        data.update(self.gps_data)
        return data.get(key)

    @property
    def data(self) -> Dict[str, Any]:
        """Raw EXIF data embedded in this image."""
        return {
            _camel_to_snake(pilx.TAGS.get(k, str(k))): v for k, v in self.__exif.items()
        }

    @property
    def gps_data(self) -> Dict[str, Any]:
        """Raw GPS data embedded in this image."""
        return {
            _camel_to_snake(pilx.GPSTAGS.get(k, str(k))): v
            for k, v in self.__exif.get_ifd(34853).items()
        }

    @property
    def make(self) -> Optional[str]:
        val: Optional[str] = self.__exif.get(271)
        return val.strip().strip("\x00") if isinstance(val, str) else None

    @property
    def model(self) -> Optional[str]:
        val: Optional[str] = self.__exif.get(272)
        return val.strip().strip("\x00") if isinstance(val, str) else None

    @property
    def x_resolution(self) -> Optional[float]:
        val: Optional[float] = self.__exif.get(282)
        return round(float(val), 4) if val is not None else None

    @property
    def y_resolution(self) -> Optional[float]:
        val: Optional[float] = self.__exif.get(283)
        return round(float(val), 4) if val is not None else None

    @property
    def software(self) -> Optional[str]:
        val: Optional[str] = self.__exif.get(305)
        return val.strip().strip("\x00") if isinstance(val, str) else None

    @property
    def date_time(self) -> Optional[dt.datetime]:
        parse = lambda v: dt.datetime.strptime(v, "%Y:%m:%d %H:%M:%S")
        val = self.__exif.get(306)
        ret = None
        try:
            ret = parse(val)
        except:
            pass
        return None if val is None else ret

    @property
    def gps_latitude(self) -> Optional[float]:
        ref = self.__exif.get_ifd(34853).get(1)
        val = self.__exif.get_ifd(34853).get(2)
        if val is None or ref is None:
            return None

        mult = -1 if ref != "N" else 1
        val = _convert_to_degrees(val) * mult
        return val

    @property
    def gps_longitude(self) -> Optional[float]:
        ref = self.__exif.get_ifd(34853).get(3)
        val = self.__exif.get_ifd(34853).get(4)
        if val is None or ref is None:
            return None

        mult = -1 if ref != "E" else 1
        val = _convert_to_degrees(val) * mult
        return val

    @property
    def gps_altitude(self) -> Optional[float]:
        val = self.__exif.get_ifd(34853).get(6)
        return None if val is None else round(float(val), 4)

    @property
    def gps_image_direction(self) -> Optional[float]:
        val = self.__exif.get_ifd(34853).get(17)
        return None if val is None else round(float(val), 4)

    @property
    def gps_map_datum(self) -> Optional[str]:
        val = self.__exif.get_ifd(34853).get(18)
        return None if not isinstance(val, str) else val.strip().strip("\x00")

    @property
    def focal_length(self) -> Optional[float]:
        val: Optional[float] = self.__exif.get(37386)
        return round(val, 4) if isinstance(val, float) else None

    @property
    def image_description(self) -> Optional[str]:
        val: Optional[str] = self.__exif.get(270)
        return val if val else None


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--file", required=True)
    args = parser.parse_args()

    cm = CameraMetadata(args.file)
    print(cm.data)
    print(cm.gps_data)
    print(cm.date_time)
    print(cm.gps_altitude)
    print(cm.gps_image_direction)
    print(cm.gps_latitude)
    print(cm.gps_longitude)
    print(cm.gps_map_datum)
    print(cm.image_description)
    print(cm.make)
    print(cm.model)
    print(cm.software)
    print(cm.x_resolution)
    print(cm.y_resolution)



"""Module containing the palette definitions flyr supports and the function
with which to use it.

Specifically see the `map_colors` function, which uses any of the palettes
defined in the `palettes` dictionary.
"""

from typing import Dict, List, Literal, Tuple, Union

import numpy as np

import flyr.normalization as norm
from flyr.dtypes import Generic2DArray, RGBArray
from flyr.palettes.cividis import cividis
from flyr.palettes.copper import copper
from flyr.palettes.gist_earth import gist_earth
from flyr.palettes.gist_rainbow import gist_rainbow
from flyr.palettes.hot import hot
from flyr.palettes.inferno import inferno
from flyr.palettes.jet import jet
from flyr.palettes.magma import magma
from flyr.palettes.ocean import ocean
from flyr.palettes.plasma import plasma
from flyr.palettes.rainbow import rainbow
from flyr.palettes.terrain import terrain
from flyr.palettes.turbo import turbo
from flyr.palettes.viridis import viridis

PaletteBuiltIn = Union[
    Literal["cividis"],
    Literal["copper"],
    Literal["gist_earth"],
    Literal["gist_rainbow"],
    Literal["hot"],
    Literal["inferno"],
    Literal["jet"],
    Literal["magma"],
    Literal["ocean"],
    Literal["plasma"],
    Literal["rainbow"],
    Literal["terrain"],
    Literal["turbo"],
    Literal["viridis"],
]
Palette = Union[
    PaletteBuiltIn,
    Literal["embedded"],
    Literal["grayscale"],
    Literal["grayscale-inverted"],
]

PaletteColors = List[Tuple[int, int, int]]

palettes: Dict[PaletteBuiltIn, List[Tuple[int, int, int]]] = {
    "cividis": cividis,
    "copper": copper,
    "gist_earth": gist_earth,
    "gist_rainbow": gist_rainbow,
    "hot": hot,
    "inferno": inferno,
    "jet": jet,
    "magma": magma,
    "ocean": ocean,
    "plasma": plasma,
    "rainbow": rainbow,
    "terrain": terrain,
    "turbo": turbo,
    "viridis": viridis,
}


def map_colors(
    thermal: Generic2DArray[np.float64],
    palette: Union[PaletteBuiltIn, PaletteColors] = "jet",
    histogram_equalized: bool = False,
) -> RGBArray:
    """Colors a normalized image array.

    Parameters
    ----------
    thermal: Generic2DArray[np.float64]
        A normalized array, that is all values are between 0 and 1.

    Returns
    -------
    RGBArray
        A 3D array with channels as its last dimension, containing integers
        between 0 and 255.
    """
    if isinstance(palette, list):
        colors = palette
    elif isinstance(palette, str):
        colors = palettes[palette]

    colored = np.zeros(thermal.shape + (3,), dtype=np.uint8)
    num_colors = len(colors)
    if histogram_equalized:
        thermal = norm.equalize_histogram(thermal, num_colors)

    for idx, color in enumerate(colors):
        lower = idx / num_colors
        upper = (idx + 1) / num_colors
        colored[(lower <= thermal) & (thermal <= upper)] = color

    return colored



from typing import Optional, Tuple

import numpy as np
from PIL import Image

import flyr.picture_in_picture as pip
from flyr.dtypes import GrayArray, MaskArray, RGBArray


def emphasize_edges(
    bg_render_pil: Image.Image,
    optical: RGBArray,
    pip_info: pip.PictureInPicture,
    edge_emphasis: float,
    color_mask: Optional[MaskArray],
    custom_shift_x: int,
    custom_shift_y: int,
) -> Image.Image:
    ratio = pip_info.real_to_ir
    offset_x = pip_info.offset_x
    offset_y = pip_info.offset_y

    ratio = optical.shape[1] / bg_render_pil.size[0] / ratio
    src_size = bg_render_pil.size
    dst_size = (
        round(bg_render_pil.size[0] * ratio),
        round(bg_render_pil.size[1] * ratio),
    )

    height, width = optical.shape[:2]
    origin_x = (width // 2) + custom_shift_x
    origin_y = (height // 2) + custom_shift_y
    origin_x = origin_x - dst_size[0] // 2 + offset_x
    origin_y = origin_y - dst_size[1] // 2 + offset_y

    thermal_surface = (
        origin_x,
        origin_y,
        origin_x + dst_size[0],
        origin_y + dst_size[1],
    )
    embossment, opacity = _embossment_masks_pil(optical, edge_emphasis)
    embossment = embossment.crop(thermal_surface).resize(src_size)
    opacity = opacity.crop(thermal_surface).resize(src_size)

    if color_mask is not None:
        embossment = _apply_mask_pil(embossment, color_mask)
        opacity = _apply_mask_pil(opacity, color_mask)

    bg_render_pil = bg_render_pil.convert("RGBA")
    bg_render_pil.paste(embossment, (0, 0), opacity)
    return bg_render_pil


def _embossment_masks(
    optical: RGBArray, opacity: float = 0.275
) -> Tuple[GrayArray, GrayArray]:
    opacity = min(max(opacity, 0.0), 1.0)

    # Define azimuth, elevation, and depth
    ele = np.pi / 2.2  # radians
    azi = np.pi / 4.0  # radians
    dep = 15.0  # (0-100)

    # Get a B&W version of the image
    img = Image.fromarray(optical).convert("L")
    arr_embossment = np.array(img).astype(float)

    # Find the gradient and adjust the gradient by the depth factor
    grad_x, grad_y = np.gradient(arr_embossment)
    grad_x = grad_x * dep / 100.0
    grad_y = grad_y * dep / 100.0

    # Get the unit incident ray
    gd = np.cos(ele)  # length of projection of ray on ground plane
    dx = gd * np.cos(azi)
    dy = gd * np.sin(azi)
    dz = np.sin(ele)

    # Find the unit normal vectors for the image
    leng = np.sqrt(grad_x**2 + grad_y**2 + 1.0)
    uni_x = grad_x / leng
    uni_y = grad_y / leng
    uni_z = 1.0 / leng

    # Take the dot product, avoiding overflow
    arr_embossment = 255 * (dx * uni_x + dy * uni_y + dz * uni_z)
    arr_embossment = arr_embossment.clip(0, 255).astype(np.uint8)

    # Threshold non-edges to be white and see-through
    arr_embossment[arr_embossment > 180] = 255
    arr_embossment = arr_embossment.astype(np.uint8)
    arr_opacity = np.full_like(arr_embossment, fill_value=round(opacity * 255))
    arr_opacity[arr_embossment > 180] = 0
    return arr_embossment, arr_opacity


def _embossment_masks_pil(
    optical: RGBArray, opacity: float = 0.275
) -> Tuple[Image.Image, Image.Image]:
    e, o = _embossment_masks(optical, opacity)
    return Image.fromarray(e), Image.fromarray(o)


def _apply_mask_pil(image: Image.Image, mask: MaskArray):
    array_image = np.asarray(image)
    masked_image = _apply_mask(array_image, mask).filled(0)
    return Image.fromarray(masked_image)


def _apply_mask(image, mask: MaskArray):
    return np.ma.masked_array(image, np.invert(mask))



import os
from distutils.core import setup

import version

# Read the contents of README.md
this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


version_nr = version.get()

setup(
    name="flyr",
    packages=["flyr", "flyr.palettes", "flyr.msx"],
    version=version_nr,
    license="EUPL v1.2",
    description="Flyr is a library for extracting thermal data from FLIR images written fully in Python, without depending on ExifTool.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Arthur Nieuwland",
    author_email="anieuwland@nimmerfort.eu",
    url="https://bitbucket.org/nimmerwoner/flyr/",
    project_urls={
        "Issues": "https://bitbucket.org/nimmerwoner/flyr/issues?status=new&status=open",
        "Releases": "https://bitbucket.org/nimmerwoner/flyr/downloads/",
        "Author website": "http://nimmerfort.eu",
    },
    download_url=f"https://bitbucket.org/nimmerwoner/flyr/downloads/flyr-{version_nr}.tar.gz",
    keywords=["flir", "thermography", "heat imagery"],
    install_requires=["numpy", "pillow"],
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    entry_points={"console_scripts": ["flyr=flyr.flyr:main"]},
)



"""Module containing the implementation of the distance estimation algorithm to
estimate the shift between thermal and optical images."""

from typing import Tuple, Union

import numpy as np
from PIL import Image, ImageOps

from flyr.dtypes import (
    GrayArray,
    MaskArray,
    RGBArray,
)
from flyr.msx.algorithms import calculate_translation
from flyr.picture_in_picture import PictureInPicture

__EPS = 1e-12  # Division by 0 prevention


def estimate_shift(
    thermal_image: GrayArray,
    optical_image: RGBArray,
    pip_info: PictureInPicture,
) -> Tuple[int, int]:
    """Estimate the shift between thermal and optical images using a distance estimation algorithm.

    Parameters
    ----------
    thermal_image: GrayArray
        The thermal image.
    optical_image: RGBArray
        The optical image.
    pip_info: PictureInPicture
        The picture in picture information.
    masks: Tuple[Optional[MaskArray], Optional[MaskArray]]
        The masks to be used in the phase correlation algorithm. If None, the masks will be
        built using the gradient edge maps of the images.

    Raises
    ------
    ValueError
        If the masks are provided but one of them is empty (contains no objects).
    ValueError
        If only one the optical mask is provided and the thermal mask is not.

    Returns
    -------
    Tuple[int, int]
        The estimated translation in the x-axis and the y-axis.
    """
    optical_crop, thermal_crop = _find_optical_thermal_matched_region(
        thermal_image, optical_image, pip_info
    )

    size = thermal_crop.shape[1], thermal_crop.shape[0]
    cover = ImageOps.cover(Image.fromarray(optical_crop), size)
    optical_crop = np.array(cover)
    thermal_edges, optical_edges = _calculate_edges(thermal_crop, optical_crop)

    if not np.any(thermal_edges):
        msg = "The mask for the thermal image must contain at least one object. But, it's empty."
        raise ValueError(msg)

    if not np.any(optical_edges):
        msg = "The mask for the optical image must contain at least one object. But, it's empty."
        raise ValueError(msg)

    return calculate_translation(optical_edges, thermal_edges)


def _calculate_edges(
    thermal_image: GrayArray,
    optical_image: RGBArray,
) -> Tuple[MaskArray, MaskArray]:
    """Generate the masks to be used in the distance estimation algorithm using
    gradient edge maps.

    Parameters
    ----------
    thermal_image: GrayArray
        The thermal image.
    optical_image: RGBArray
        The optical image.
    masks: Tuple[Optional[MaskArray], Optional[MaskArray]]
        The masks to be used in the phase correlation algorithm.

    Raises
    ------
    ValueError
        If only the optical mask is provided and the thermal mask is not.

    Returns
    -------
    Tuple[MaskArray, MaskArray]
        The masks to be used in the phase correlation algorithm.
    """
    thermal_mask = _build_gradient_edge_maps(thermal_image)
    optical_mask = _build_gradient_edge_maps(optical_image)
    return thermal_mask, optical_mask


def _find_optical_thermal_matched_region(
    thermal: GrayArray, optical: RGBArray, pip_info: PictureInPicture
) -> Tuple[RGBArray, GrayArray]:
    """
    Finds the optical and thermal images scene regions aligned to the thermal PiP
    crop box.

    Parameters
    ----------
    thermal: GrayArray
        The thermal image.
    optical: RGBArray
        The optical image.
    pip_info: PictureInPicture
        The picture in picture information.

    Returns
    -------
    Tuple[RGBArray, GrayArray]
        The cropped optical and thermal images aligned to the thermal PiP crop box.
    """
    # Compute optical crop bounds using picture_in_picture_pil math.
    x1, y1, x2, y2 = pip_info.crop_box
    ratio = pip_info.real_to_ir
    render_w = thermal.shape[1]
    ratio = optical.shape[1] / render_w / ratio
    dst_w = round((x2 - x1) * ratio)
    dst_h = round((y2 - y1) * ratio)

    width = optical.shape[1]
    height = optical.shape[0]
    origin_x = width // 2 - dst_w // 2 + pip_info.offset_x
    origin_y = height // 2 - dst_h // 2 + pip_info.offset_y

    optical_pil = Image.fromarray(optical)
    optical_crop = optical_pil.crop(
        (origin_x, origin_y, origin_x + dst_w, origin_y + dst_h)
    )

    return np.array(optical_crop), thermal


def _build_gradient_edge_maps(image: Union[GrayArray, RGBArray]) -> MaskArray:
    """Build the gradient edge map of the image using a thresholding method.

    An L1 normalization is applied to the gradients to amplify edges.

    Parameters
    ----------
    image : Union[GrayArray, RGBArray]
        The image to build the gradient edge map from.

    Returns
    -------
    MaskArray
        The gradient edge map of the image.
    """
    thresholded_edges = _otsu_thresold(image)
    gradients = np.gradient(thresholded_edges.astype(np.float32))
    gradient_magnitude = np.sum(np.abs(gradients), axis=0)
    return gradient_magnitude.astype(np.uint8) > 0


def _otsu_thresold(image: Union[GrayArray, RGBArray]) -> GrayArray:
    """Applies Otsu's thresholding method to the input image.

    Parameters
    ----------
    image: Union[GrayArray, RGBArray]
        The input image to be thresholded, either grayscale or RGB which will
        then be converted to grayscale.

    Returns
    -------
    GrayArray
        The thresholded image.
    """
    image = np.array(Image.fromarray(image).convert("L"))
    if image.size == 0:
        return image.astype(np.uint8)

    min_val, max_val = float(image.min()), float(image.max())
    hist, bin_edges = np.histogram(image.ravel(), bins=256, range=(min_val, max_val))
    hist = hist.astype(np.float64)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.0

    prob = hist / hist.sum()
    cum_prob = np.cumsum(prob)
    cum_mean = np.cumsum(prob * bin_centers)
    global_mean = cum_mean[-1]

    between_var = (global_mean * cum_prob - cum_mean) ** 2 / (
        cum_prob * (1.0 - cum_prob) + __EPS
    )
    threshold = bin_centers[int(np.argmax(between_var))]

    return ~((image > threshold).astype(np.uint8) * 255)
